# -*- coding: utf-8 -*-

import logging


logger = logging.getLogger(__name__)


def get_suite_class(options):
    if not options.app_processes and options.gevent_greanlets:
        from noseapp.suite.bases.gevent_suite import GeventSuite
        return GeventSuite

    elif options.app_processes and options.thread_pool:
        from noseapp.suite.bases.threading_suite import ThreadSuite
        return ThreadSuite

    from noseapp.suite.bases.simple import BaseSuite
    return BaseSuite


class TestCaseMediator(object):
    """
    Mediator class between noseapp.suite.Suite, TestLoader, TestSuite,
    TestCase. Provides between TestCase->TestLoader->TestSuite.
    """

    def __init__(self, require):
        self._test_cases = []
        self._require = require

    @property
    def test_cases(self):
        return self._test_cases

    def create_map(self):
        """
        {
            'class name': {
                'cls': 'link to class object',
                'tests': {
                    'method name': 'link to class method',
                },
            },
        }

        :return: dict
        """
        mp = {}

        for case in self._test_cases:
            mp[case.__name__] = {
                'cls': case,
                'tests': dict(
                    (atr, getattr(case, atr))
                    for atr in dir(case)
                    if atr.startswith('test')
                    or
                    atr.startswith('runTest'),
                ),
            }

        return mp

    def create_suite(self, nose_config, test_loader):
        """
        Create suite instance
        """
        suite_class = get_suite_class(nose_config.options)
        suite = suite_class(config=nose_config)

        for case in self._test_cases:
            suite.addTests(
                test_loader.loadTestsFromTestCase(
                    case.with_require(self._require),
                ),
            )

        return suite

    def add_test_case(self, test_case):
        self._test_cases.append(test_case)



# -*- coding: utf-8 -*-

from unittest import TestCase


class TestPushApp(TestCase):
    """
    Test push app with call plugin instance
    """

    def runTest(self):
        from noseapp import AppPlugin

        plugin = AppPlugin()
        self.assertFalse(hasattr(plugin, 'app'))

        instance = plugin(object())
        self.assertIsInstance(instance, AppPlugin)
        self.assertTrue(hasattr(plugin, 'app'))
        self.assertIsInstance(plugin.app, object)



# -*- coding: utf-8 -*-

import time
from contextlib import contextmanager

from nose.core import TextTestRunner as BaseTestRunner


@contextmanager
def measure_time(result):
    start = time.time()
    yield
    stop = time.time()

    result.printErrors()
    result.printSummary(start, stop)


__all__ = (
    BaseTestRunner,
)



# -*- coding: utf-8 -*-

"""
NoseApp management.

Command line interface:
    noseapp-manage [command, command args, command kwargs]

Example:
    noseapp-manage command argument1 argument2 argument3=value
"""

import os
import imp
import sys
import inspect
from importlib import import_module
from collections import OrderedDict


_COMMANDS = OrderedDict()


def _error(msg):
    sys.stdout.write('error: {}\n'.format(msg))
    sys.exit(1)


def _get_command_name():
    try:
        return sys.argv[1]
    except IndexError:
        _error('command name is required')


def _get_command(name):
    try:
        return _COMMANDS[name]
    except KeyError:
        _error('command "{}" is not found'.format(name))


def _get_create_app_func(path):
    try:
        module, func = path.split(':')
    except ValueError:
        _error('incorrect path to importing of getting application function')
    except IndexError:
        _error('import path is required as first argument of signature')

    module = import_module(module)
    func = getattr(module, func, None)

    if callable(func):
        return func

    _error('Function for getting application is not found')


def _load_manage_py():
    manage_py = os.path.join(os.getcwd(), 'manage.py')

    if os.path.isfile(manage_py):
        manage_module = imp.new_module(manage_py.rstrip('.py'))
        manage_module.__file__ = manage_py

        try:
            execfile(manage_py, manage_module.__dict__)
        except IOError as e:
            e.strerror = 'Unable to load manage file "{}"'.format(e.strerror)
            raise


def register_command(name, command):
    assert callable(command), 'command is not callable'
    assert isinstance(name, basestring), 'name of command must be string only'

    _COMMANDS[name] = command


class DefaultCommands(object):

    @staticmethod
    def help(command=None, **kw):
        """
        :app_path: path to importing of getting application function
        """
        sys.stdout.write(__doc__)

        if command:
            command_func = _get_command(command)

            sys.stdout.write('\nCommand: {} ({}):\n'.format(command, inspect.getfile(command_func)))

            sys.stdout.write('{}\n'.format(command_func.__doc__))
        else:
            sys.stdout.write('\nCommands:\n\n')

            for command_name in _COMMANDS:
                sys.stdout.write('    * {}\n'.format(command_name))

        sys.stdout.write('\n')

    @staticmethod
    def run_app(app_path, **kwargs):
        """
        Command for running application.

        :usage:
            noseapp-manage run import.path.to:get_app_function

        :app_path: path for importing application
        :kwargs: kwargs of create application function
        """
        for argv in sys.argv:
            if '--gevent' == argv:
                from gevent.monkey import patch_all
                patch_all()
                break

        app = _get_create_app_func(app_path)(**kwargs)
        app.run()

    @staticmethod
    def run_many(*app_path):
        """
        Command for running many applications

        :usage:
            noseapp-manage run-many import.path.to:get_app_one import.path.to:get_app_two
        """
        exit_code = 0

        for path in app_path:
            try:
                DefaultCommands.run_app(path)
            except SystemExit as e:
                if e.code > exit_code:
                    exit_code = e.code

        sys.exit(exit_code)

    @staticmethod
    def create_project(name):  # TODO
        """
        Create base structure of project
        """
        pass


# Add commands here. Order is important.
register_command('help', DefaultCommands.help)
register_command('run', DefaultCommands.run_app)
register_command('run_many', DefaultCommands.run_many)
register_command('create_project', DefaultCommands.create_project)


def run():
    _load_manage_py()

    command = _get_command(
        _get_command_name(),
    )

    args = []
    kwargs = {}

    arguments = []

    for arg in sys.argv[2:]:
        if arg.startswith('-'):
            break

        arguments.append(arg)

    for arg in arguments:
        if '=' in arg:
            k, v = arg.split('=')
            kwargs[k.strip()] = v.strip()
        else:
            args.append(arg.strip())

    try:
        command(*args, **kwargs)
    except TypeError:
        _error('signature of command is invalid')


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-

import time


class TimeoutException(BaseException):
    pass


def waiting_for(func, timeout=30, sleep=0.5, args=None, kwargs=None):
    """
    :param func: callback function
    :param timeout: number of seconds for timeout
    :param sleep: number of second for sleep after call function
    :param args, kwargs: callback arguments
    """
    args = args or tuple()
    kwargs = kwargs or {}

    t_start = time.time()

    while time.time() <= t_start + timeout:
        result = func(*args, **kwargs)

        if result:
            return result

        time.sleep(sleep)
    else:
        mess = 'Timeout {timeout} exceeded.'.format(timeout=timeout)

        raise TimeoutException(mess)



# -*- coding: utf-8 -*-

from threading import Lock
from threading import Thread
from Queue import Queue as TaskQueue

from noseapp.utils.common import waiting_for
from noseapp.runner.base import measure_time
from noseapp.runner.base import BaseTestRunner
from noseapp.utils.common import TimeoutException


Process = ResultQueue = cpu_count = None


def _import_mp():
    """
    Import multiprocessing
    """
    global Process, ResultQueue, cpu_count

    from multiprocessing import Process
    from multiprocessing import Manager
    from multiprocessing import cpu_count

    m = Manager()

    Process, ResultQueue, cpu_count = (
        Process, m.Queue, cpu_count
    )


def task(suite, result, result_queue):
    """
    Task to perform suite
    """
    suite(result)

    del suite

    def get_value(value):
        if isinstance(value, basestring):
            return value
        return repr(value)

    failures = [[get_value(v) for v in f] for f in result.failures]
    errors = [[get_value(v) for v in e] for e in result.errors]
    skipped = [[get_value(v) for v in s] for s in result.skipped]

    result_queue.put_nowait([failures, errors, skipped, result.testsRun])


def run(processor, queue_handler):
    """
    Go, go, go!
    """
    workers = [
        Thread(target=processor.serve),
        Thread(target=queue_handler.handle),
    ]

    try:
        for worker in workers:
            worker.start()
        for worker in workers:
            worker.join()
    except KeyboardInterrupt:
        processor.destroy()
    finally:
        processor.close()


class ResultQueueHandler(object):

    TEST_REPR = 'Test({})'

    def __init__(self, suites, result, result_queue):
        self._result = result
        self._comparison = {}
        self._result_queue = result_queue

        self._counter = 0
        self._length_suites = 0

        self._match(suites)

        self._counter_lock = Lock()

    def _match(self, suites):
        """
        Create dict: repr -> test
        """
        suite_list = [s for s in suites]
        self._length_suites = len(suite_list)

        for suite in suite_list:
            for test in suite:
                key = self.TEST_REPR.format(
                    repr(test.test),
                )

                if key in self._comparison:
                    raise ValueError('Test __repr__ "{}" already exist'.format(key))

                self._comparison[key] = test.test

    def _create_result(self, data):
        _repr, messages = data[0], data[1:]
        test = self._comparison[_repr]
        del self._comparison[_repr]
        return tuple([test] + messages)

    def _add_failures(self, failures):
        for fail in failures:
            self._result.failures.append(
                self._create_result(fail),
            )

    def _add_errors(self, errors):
        for err in errors:
            self._result.failures.append(
                self._create_result(err),
            )

    def _add_skipped(self, skipped):
        for skip in skipped:
            self._result.failures.append(
                self._create_result(skip),
            )

    def handle(self):
        while self._counter < self._length_suites:
            failures, errors, skipped, tests_run = self._result_queue.get()

            self._add_failures(failures)
            self._add_errors(errors)
            self._add_skipped(skipped)

            with self._counter_lock:
                self._result.testsRun += tests_run
                self._counter += 1


class TaskProcessor(object):
    """
    Collect tasks and run with multiprocessing.Process
    """

    def __init__(self, processes, process_timeout=1800):
        """
        :param processes: nun processes
        :type processes: int
        :param process_timeout: number of seconds for process timeout
        :type process_timeout: int, float
        """
        self._processes = processes if processes > 0 else cpu_count()
        self._current = []
        self._queue = TaskQueue()
        self._process_timeout = process_timeout

    def _is_release(self):
        if len(self._current) < self._processes:
            return True

        for process in self._current:
            if not process.is_alive():
                process.terminate()
                self._current.remove(process)
                return True

        return False

    def add_task(self, target, args=None, kwargs=None):
        self._queue.put_nowait((target, args or tuple(), kwargs or dict()))

    def serve(self):
        while not self._queue.empty():
            try:
                waiting_for(self._is_release, timeout=self._process_timeout, sleep=0.01)
            except TimeoutException:
                self.destroy()
                raise

            target, args, kwargs = self._queue.get_nowait()
            process = Process(target=target, args=args, kwargs=kwargs)
            process.start()
            self._current.append(process)

    def destroy(self):
        """
        Kill all processes
        """
        for process in self._current:
            process.terminate()

    def close(self):
        for process in self._current:
            process.join()


class MultiprocessingTestRunner(BaseTestRunner):
    """
    Run suites with multiprocessing.Process
    """

    def run(self, suites):
        wrapper = self.config.plugins.prepareTest(suites)
        if wrapper is not None:
            suites = wrapper

        wrapped = self.config.plugins.setOutputStream(self.stream)
        if wrapped is not None:
            self.stream = wrapped

        result = self._makeResult()

        _import_mp()

        result_queue = ResultQueue()
        processor = TaskProcessor(self.config.options.app_processes)
        queue_handler = ResultQueueHandler(suites, result, result_queue)

        for suite in suites:
            processor.add_task(task, args=(suite, result, result_queue))

        with measure_time(result):
            run(processor, queue_handler)

        self.config.plugins.finalize(result)

        return result



# -*- coding: utf8 -*-

import sys

from nose.core import TestProgram as BaseTestProgram

from noseapp.core import extensions
from noseapp.core.collector import CollectSuite


def get_test_runner_class(options):
    """
    :return: nose.core.TextTestRunner
    """
    if options.app_processes:
        from noseapp.runner.multiprocessing_runner import MultiprocessingTestRunner
        return MultiprocessingTestRunner

    elif options.gevent_pool:
        from noseapp.runner.gevent_runner import GeventTestRunner
        return GeventTestRunner

    elif options.thread_pool:
        from noseapp.runner.threading_runner import ThreadingTestRunner
        return ThreadingTestRunner

    from noseapp.runner.base import BaseTestRunner
    return BaseTestRunner


class TestProgram(BaseTestProgram):

    collector_class = CollectSuite

    def __init__(self, app, argv, *args, **kwargs):
        self._app = app
        self._argv = argv

        super(TestProgram, self).__init__(argv=argv, *args, **kwargs)

        self._app.config.init_nose_config(self.config)
        self._app.initialize()

    # disarm base class
    def createTests(self):
        pass

    def runTests(self):
        pass

    def perform(self):
        """
        Perform test program
        """
        collector = self.collector_class(
            self._argv, self._app.suites, self.testLoader, self.config,
        )
        suites = collector.make_result()

        extensions.clear()

        test_runner = get_test_runner_class(self.config.options)

        self.testRunner = test_runner(stream=self.config.stream,
                                      verbosity=self.config.verbosity,
                                      config=self.config)

        plug_runner = self.config.plugins.prepareTestRunner(self.testRunner)

        if plug_runner is not None:
            self.testRunner = plug_runner

        self._app.before()

        result = self.testRunner.run(suites)
        self.success = result.wasSuccessful()

        self._app.after()

        if self.exit:
            sys.exit(not self.success)

        return self.success



# -*- coding: utf-8 -*-

from noseapp.case.base import TestCase
from noseapp.case.screenplay import step
from noseapp.case.screenplay import ScreenPlayCase


__all__ = (
    step,
    TestCase,
    ScreenPlayCase,
)



# -*- coding: utf8 -*-

from multiprocessing import cpu_count
from multiprocessing.pool import ThreadPool

from noseapp.runner.base import measure_time
from noseapp.runner.base import BaseTestRunner


class ThreadingTestRunner(BaseTestRunner):
    """
    Run tests with multiprocessing.pool.ThreadPool
    """

    def run(self, suites):
        wrapper = self.config.plugins.prepareTest(suites)
        if wrapper is not None:
            suites = wrapper

        wrapped = self.config.plugins.setOutputStream(self.stream)
        if wrapped is not None:
            self.stream = wrapped

        result = self._makeResult()

        size = self.config.options.thread_pool
        if size < 0:
            size = cpu_count()

        pool = ThreadPool(size)

        with measure_time(result):

            for suite in suites:
                pool.apply_async(suite, args=(result,))

            pool.close()
            pool.join()

        self.config.plugins.finalize(result)
        return result



# -*- coding: utf-8 -*-

from unittest import TestCase


class TestModifyDict(TestCase):
    """
    Test noseapp.datastructures.ModifyDict class
    """

    def runTest(self):
        from noseapp.datastructures import ModifyDict

        dct = ModifyDict(test=1)

        self.assertIn('test', dct)
        self.assertEqual(dct['test'], 1)
        self.assertEqual(dct.get('none'), None)

        dct.hello = 'world'

        self.assertEqual(dct.hello, 'world')

        self.assertRaises(AttributeError, lambda: dct.no_attr)



# -*- coding: utf-8 -*-

import re

from noseapp.core.collector import console
from noseapp.suite.bases.simple import BaseSuite


CASE_COLLECT_STRATEGY = 'case'
BASIC_COLLECT_STRATEGY = 'basic'
SUITE_COLLECT_STRATEGY = 'suite'
METHOD_COLLECT_STRATEGY = 'method'

COLLECTOR_COMMAND_PATTERN = re.compile(r'^.*\.|^.*:')
COLLECT_CASE_COMMAND_PATTERN = re.compile(r'^.*\:')
COLLECT_METHOD_COMMAND_PATTERN = re.compile(r'^.*\:.*\..*')


class CollectError(BaseException):
    pass


def get_collector_strategy(command):
    if command and COLLECTOR_COMMAND_PATTERN.search(command) is not None:

        if COLLECT_METHOD_COMMAND_PATTERN.search(command) is not None:
            return METHOD_COLLECT_STRATEGY

        if COLLECT_CASE_COMMAND_PATTERN.search(command) is not None:
            return CASE_COLLECT_STRATEGY

        return SUITE_COLLECT_STRATEGY

    return BASIC_COLLECT_STRATEGY


def get_suite_by_name(name, suites):
    for suite in suites:

        if suite.name == name:
            return suite

    else:
        raise CollectError(
            'Suite "{}" is not found'.format(name),
        )


def get_case_from_suite(case_name, suite):
    try:
        return suite.get_map()[case_name]['cls']
    except KeyError:
        raise CollectError('TestCase "{}" is not found'.format(case_name))


class CollectSuite(object):
    """
    Collect suite for test runner
    """

    def __init__(self, argv, suites, test_loader, nose_config):
        if nose_config.options.ls:
            console.tree(suites, show_docs=nose_config.options.doc)

        self._suites = suites
        self._nose_config = nose_config
        self._test_loader = test_loader

        if nose_config.options.run_test:
            self._command = nose_config.options.run_test
        elif len(argv) > 2:
            self._command = argv[1]
        else:
            self._command = None

    @property
    def command(self):
        return self._command

    @property
    def suites(self):
        return self._suites

    @property
    def collect(self):  # syntax of sugar :)
        strategy_to_method = {
            CASE_COLLECT_STRATEGY: self._collect_by_case_strategy,
            BASIC_COLLECT_STRATEGY: self._collect_by_basic_strategy,
            SUITE_COLLECT_STRATEGY: self._collect_by_suite_strategy,
            METHOD_COLLECT_STRATEGY: self._collect_by_method_strategy,
        }

        return strategy_to_method[get_collector_strategy(self._command)]

    def make_result(self):
        return self._test_loader.suiteClass(
            self.collect(),
        )

    def _collect_by_basic_strategy(self):
        return [
            suite(self._nose_config, self._test_loader)
            for suite in self._suites
        ]

    def _collect_by_suite_strategy(self):
        suite = get_suite_by_name(self._command, self._suites)

        return [suite(self._nose_config, self._test_loader)]

    def _collect_by_case_strategy(self):
        suite_name, case_name = self._command.split(':')

        app_suite = get_suite_by_name(suite_name, self._suites)
        app_suite.init_extensions()

        case = get_case_from_suite(case_name, app_suite)

        suite = BaseSuite(config=self._nose_config)
        suite.addTests(
            self._test_loader.loadTestsFromTestCase(case),
        )

        return [suite]

    def _collect_by_method_strategy(self):
        suite_name, case_info = self._command.split(':')
        case_name, method_name = case_info.split('.')

        app_suite = get_suite_by_name(suite_name, self._suites)
        app_suite.init_extensions()

        case = get_case_from_suite(case_name, app_suite)

        try:
            suite = BaseSuite(
                map(case, [method_name]),
                config=self._nose_config,
            )
        except ValueError:
            raise CollectError(
                'Method "{}" of class "{}" is not found'.format(method_name, case_name)
            )

        return [suite]



# -*- coding: utf8 -*-

from noseapp.core.program import TestProgram
from noseapp.core.collector import CollectSuite
from noseapp.core.loader import load_suites_from_path


__all__ = (
    TestProgram,
    CollectSuite,
    load_suites_from_path,
)



# -*- coding: utf-8 -*-

from noseapp.suite.mediator import TestCaseMediator


class Suite(object):
    """
    Base Suite class for group or one TestCase.

    Usage DEFAULT_REQUIRE property for default extensions require.
    """

    mediator_class = TestCaseMediator

    def __init__(self, name, require=None):
        """
        :param name: suite name
        :type name: str
        :param require: extension names list
        :type require: list
        """
        self._name = name

        if hasattr(self, 'DEFAULT_REQUIRE'):
            self._require = self.DEFAULT_REQUIRE + (require or [])
        else:
            self._require = require

        self._mediator = self.mediator_class(require=self._require)

    @property
    def name(self):
        return self._name

    @property
    def require(self):
        return self._require

    def register(self, cls):
        """
        Add test case class

        :type cls: noseapp.case.TestCase
        """
        self._mediator.add_test_case(cls)
        return cls

    def get_map(self):
        return self._mediator.create_map()

    def init_extensions(self):
        """
        Init extensions for test cases. Without build suite.
        """
        for case in self._mediator.test_cases:
            case.with_require(self._require)

    def __call__(self, nose_config, test_loader):
        """
        Build suite
        """
        return self._mediator.create_suite(nose_config, test_loader)

    def __repr__(self):
        return '<Suite {}>'.format(self._name)



# -*- coding: utf8 -*-

from nose.suite import ContextSuite


class BaseSuite(ContextSuite):
    """
    Base suite
    """

    def _run(self, result, orig):
        for test in self._tests:
                if result.shouldStop:
                    break
                test(orig)

    def run(self, result):
        if self.resultProxy:
            result, orig = self.resultProxy(result, self), result
        else:
            result, orig = result, result

        try:
            self.setUp()
        except KeyboardInterrupt:
            raise
        except:
            self.error_context = 'setup'
            result.addError(self, self._exc_info())
            return

        try:
            self._run(result, orig)
        finally:
            self.has_run = True

            try:
                self.tearDown()
            except KeyboardInterrupt:
                raise
            except:
                self.error_context = 'teardown'
                result.addError(self, self._exc_info())



# -*- coding: utf-8 -*-

from unittest import TestCase

from noseapp import NoseApp as _NoseApp


class NoseApp(_NoseApp):

    def initialize(self):
        pass


class TestInitExtensions(TestCase):
    """
    Test init extension without share
    """

    @staticmethod
    def create_fake_extension():

        class FakeExtension(object):

            name = 'test'

        return FakeExtension

    def runTest(self):
        from noseapp import TestCase
        from noseapp.core.extensions import ExtensionNotFound

        app = NoseApp()
        ex_cls = self.create_fake_extension()
        app.shared_extension(cls=ex_cls)

        TestCase.with_require(require=['test'])

        self.assertTrue(hasattr(TestCase, 'test'))
        self.assertIsInstance(TestCase.test, ex_cls)
        self.assertRaises(ExtensionNotFound, TestCase.with_require, require=['no_ex'])


class TestScreenPlayCase(TestCase):
    """
    Test noseapp.case.ScreenPlayCase class
    TODO: Test flows
    """

    @staticmethod
    def create_test_class():
        from noseapp.case import step
        from noseapp import ScreenPlayCase

        class StepByStep(ScreenPlayCase):

            def begin(self):
                self.counter = 0

            @step(1)
            def step_one(self):
                self.counter += 1

            @step(2)
            def step_two(self):
                self.counter += 1

            def finalize(self):
                self.counter += 1

        return StepByStep

    def runTest(self):
        from noseapp.case.screenplay import ScreenPlayCaseMeta

        case_cls = self.create_test_class()

        self.assertEqual(case_cls.FLOWS, None)
        self.assertEqual(case_cls.USE_PROMPT, False)
        self.assertEqual(case_cls.__metaclass__, ScreenPlayCaseMeta)

        case = case_cls(methodName='runTest')

        self.assertTrue(hasattr(case, 'runTest'))

        case.runTest()

        self.assertEqual(case.counter, 3)



# -*- coding: utf-8 -*-


from noseapp.suite import Suite
from noseapp.app import NoseApp
from noseapp.case import TestCase
from noseapp.plugins import AppPlugin
from noseapp.case import ScreenPlayCase


__all__ = (
    Suite,
    NoseApp,
    TestCase,
    AppPlugin,
    ScreenPlayCase,
)



# -*- coding: utf8 -*-

"""
=================
Step by step case
=================

Example::

  class CaseExample(ScreenPlayCase):

    USE_PROMPT = True  # usage interactive debug

    @step(1, 'description')
    def step_1(self):
      self.assertTrue(True)

    @step(2, 'description')
    def step_2(self):
      self.assertTrue(True)


  class CaseParametrizeExample(ScreenPlayCase):

    FLOWS = (
      1, 2, 3
    )

    @step(1, u'Step 1')
    def step_1(self, i):
      self.assertGreater(i, 0)

    @step(2, u'Step 2')
    def step_2(self, i):
      self.assertGreater(i, 0)
"""

import re
import sys
import logging
import traceback
from functools import wraps

from noseapp.case.base import TestCase


logger = logging.getLogger(__name__)


DOC_ATTRIBUTE_NAME = '__DOC__'
WEIGHT_ATTRIBUTE_NAME = '__WEIGHT__'
SCREENPLAY_ATTRIBUTE_NAME = '__SCREENPLAY__'

STEP_INFO_PATTERN = u'<{}.{}> Step {} "{}"'

ERROR_INFO_PATTERN = u"""

* {traceback}

* History:
{history}

* Point:
{case}.{method} -> Step {step} "{step_doc}"

* Flow:
{flow}

* Raised:
{raised}

* Message:
{message}
"""

EXCLUDE_METHOD_PATTERN = re.compile(
    r'^fail[A-Z]|^fail$|^assert[A-Z]|^assert_$',
)


class StepError(BaseException):
    pass


class StepFail(AssertionError):
    pass


def step(num, doc=''):
    """
    Decorator. Use wrapper method as case step.

    :param num: step num
    :type num: int
    :param doc: step description
    :type doc: str, unicode
    """

    def wrapper(f):
        if not isinstance(num, int):
            raise ValueError('step num mast be int only')

        setattr(f, DOC_ATTRIBUTE_NAME, doc)
        setattr(f, WEIGHT_ATTRIBUTE_NAME, num)
        setattr(f, SCREENPLAY_ATTRIBUTE_NAME, '')

        @wraps(f)
        def wrapped(*args, **kwargs):
            return f(*args, **kwargs)

        return wrapped

    return wrapper


def get_step_info(case, method):
    """
    :param case: TestCase instance
    :param method: step method

    :return: tuple
    """
    case_name = case.__class__.__name__
    method_name = method.__func__.__name__
    weight = getattr(method, WEIGHT_ATTRIBUTE_NAME)
    doc = getattr(method, DOC_ATTRIBUTE_NAME)

    return case_name, method_name, weight, doc


def perform_prompt(case_name, method_name, exit_code=0):
    """
    Interactive mode for debug

    :param case_name: case class name
    :param method_name: step method name
    :param exit_code: exit code
    """
    commands = {  # prompt commands
        'exit': 'q',
        'debug': 'd',
        'continue': 'c',
    }
    prompt = '>> {}.{} [{}]: '.format(
        case_name, method_name, ', '.join([c for _, c in commands.items()])
    )
    command = raw_input(prompt).strip()

    if command == commands['exit']:
        sys.exit(exit_code)
    elif command == commands['debug']:
        try:
            import ipdb as pdb
        except ImportError:
            import pdb

        pdb.set_trace()


def error_handler(e, msg):
    """
    Handler for raised exception
    """
    if isinstance(e, AssertionError):
        raise StepFail(msg)

    raise StepError(msg)


def run_step(case, method, flow=None):
    """
    Run step

    :param case: TestCase instance
    :param method: step method
    :param flow: from FLOWS property
    """
    case_name, method_name, weight, doc = get_step_info(case, method)
    step_info = STEP_INFO_PATTERN.format(
        case_name, method_name, weight, doc,
    )

    if case.USE_PROMPT:
        perform_prompt(case_name, method_name)

    logger.info(step_info)

    try:
        if flow is not None:
            method(case, flow)
        else:
            method(case)
    except BaseException as e:
        orig_tb = traceback.format_exc()
        history = u'\n'.join(case.__history)
        exc_cls_name = e.__class__.__name__

        try:
            msg = ERROR_INFO_PATTERN.format(
                history=history,
                case=case_name,
                method=method_name,
                step=weight,
                step_doc=doc,
                flow=flow,
                raised=exc_cls_name,
                message=e.message,
                traceback=orig_tb,
            ).encode('utf-8', 'replace')
        except UnicodeDecodeError:
            msg = ERROR_INFO_PATTERN.format(
                history=history,
                case=case_name,
                method=method_name,
                step=weight,
                step_doc=doc,
                flow=flow,
                raised=exc_cls_name,
                message=e.message,
                traceback=orig_tb.decode('utf8'),
            ).encode('utf-8', 'replace')

        error_handler(e, msg)


def make_run_test(steps):
    """
    Create runTest method

    :param steps: steps list
    """

    def run_test(self):
        run_test.__doc__ = self.__doc__

        self.begin()

        self.__history = []
        history_line = u'{}. {}'

        if self.FLOWS and hasattr(self.FLOWS, '__iter__'):

            for flow in self.FLOWS:
                for step_method in steps:
                    _, _, step, doc = get_step_info(self, step_method)
                    self.__history.append(history_line.format(step, doc))

                    run_step(self, step_method, flow=flow)
                else:
                    self.__history = []

        else:

            for step_method in steps:
                _, _, step, doc = get_step_info(self, step_method)
                self.__history.append(history_line.format(step, doc))

                run_step(self, step_method)

        self.finalize()

    return run_test


class ScreenPlayCaseMeta(type):
    """
    Build step methods and create runTest
    """

    def __new__(cls, name, bases, dct):
        new_cls = type.__new__(cls, name, bases, dct)

        steps = []

        attributes = (
            a for a in dir(new_cls)
            if not a.startswith('_')
            and
            EXCLUDE_METHOD_PATTERN.search(a) is None
        )

        for atr in attributes:
            method = getattr(new_cls, atr, None)
            if hasattr(method, SCREENPLAY_ATTRIBUTE_NAME):
                steps.append(method)

        if steps:
            steps.sort(
                key=lambda m: getattr(m, WEIGHT_ATTRIBUTE_NAME),
            )
            new_cls.runTest = make_run_test(steps)

        return new_cls


class ScreenPlayCase(TestCase):
    """
    Base TestCase class. Enjoy :)
    """

    __metaclass__ = ScreenPlayCaseMeta

    FLOWS = None
    USE_PROMPT = False

    def begin(self):
        """
        Call before run steps
        """
        pass

    def finalize(self):
        """
        Call after run steps.
        If exception in step method be raised,
        this method can not be called.
        """
        pass
